import time
import numpy as np
from skimage.transform import rescale
from skimage import io
from skimage import draw
import datetime
import json
import random
from do_every import do_every


# Make the path find the Aravis Library
# try:
#     dir_path = os.path.dirname(os.path.realpath(__file__))
#     libPath = Path(dir_path).parent.parent.joinpath('lib/aravis_arm64')
#     if os.getenv("GI_TYPELIB_PATH"):
#         os.environ["GI_TYPELIB_PATH"] = os.getenv("GI_TYPELIB_PATH") + os.pathsep + str(libPath)
#     else:
#         os.environ["GI_TYPELIB_PATH"] = str(libPath)
#     print(libPath, os.environ.get("GI_TYPELIB_PATH"))
# except:
#     print('darn')

import gi
gi.require_version('Aravis', '0.8')
from gi.repository import Aravis  # nopep8
from pathlib import Path


# Aravis.enable_interface("Fake")

currentShutter = None
currentDelay = None
currentGain = None
currentexposure = None
lastPictureTime = 0.0
lastPicture = None
background = None
allowTrigger = True
softwareTriggerFPS = -1 # 0 or below is off
frameCount = 0
gStream = None
imgShape = None


def initCam():
    # try:
    global cam, gStream, imgShape, softwareTriggerFPS
    print('aqcuiring camera')
    cam = Aravis.Camera.new(None)
    print('initializing camera')
    # cam.init()  # Initialize camera
    cam.clear_triggers()
    print('trigger mode:', cam.get_trigger_source())
    try:
        cam.AcquisitionMode = 'Continuous'
        print('set Acquisition mode to Continuous')
    except:
        print('failed to set AcquisitionMode. Mode is currently ',
              cam.get_acquisition_mode().value_name)
    # TODO
    #cam.AcquisitionFrameRateEnable = False


    cam.set_gain_auto(Aravis.Auto.OFF)
    cam.set_exposure_time_auto(Aravis.Auto.OFF)
    cam.set_frame_rate(30.0)
    if cam.get_vendor_name() == 'Allied Vision':
        cam.set_frame_rate(30.0)
    cam.set_pixel_format(Aravis.PIXEL_FORMAT_RGB_8_PACKED)
    imgShape = {"width": cam.get_width_bounds(
    ).max, "height": cam.get_height_bounds().max}
    cam.set_region(0, 0, width=imgShape["width"], height=imgShape["height"])

    setDefaults(cam)
    print("opening Camera")
    payload = cam.get_payload()
    gStream = cam.create_stream(None, None)
    gStream.push_buffer(Aravis.Buffer.new_allocate(payload))
    cam.start_acquisition()
    print('software trigger fps', softwareTriggerFPS)
    if softwareTriggerFPS > 0:
        print('going to software trigger')
        cam.set_trigger("Software")
        def callback(args):
            cam2 = args[0]
            cam2.software_trigger()
        do_every(callback, softwareTriggerFPS, cam)
    print(getInfo(cam))
    return cam


def setDefaults(cam):
    cam.set_exposure_time_auto(Aravis.Auto.OFF)
    cam.set_gain_auto(Aravis.Auto.OFF)
    cam.set_exposure_mode(Aravis.ExposureMode.TIMED)


def restartCamera(cam0):
    print("restart camera")
    cam.stop_acquisition()
    cam.start_acquisition()


def getInfo(cam):
    [x, y, width, height] = cam.get_region()
    vendor = str(cam.get_vendor_name())
    model = str(cam.get_model_name())
    result = (f"vendor: {vendor} \n"
              f"model: {model} \n"
              f"region: {x}, {y}, {width}, {height} \n"
              f"exposure: {cam.get_exposure_time()} \n"
              f"frame rate: {cam.get_frame_rate()} \n"
              f"gain: {cam.get_gain()} \n"
              f"frame rate bounds: {cam.get_frame_rate_bounds()} \n"
              f"exposure bounds: {cam.get_exposure_time_bounds()} \n"
              f"gain bounds: {cam.get_gain_bounds()} \n"
              )
    return result

def getExposureTime(cam):
    range = cam.get_exposure_time_bounds()
    return {"value":cam.get_exposure_time(), "min":range.min, "max":range.max}

def getFrameRate(cam):
    range = cam.get_frame_rate_bounds()
    return {"value":cam.get_frame_rate(), "min":range.min, "max":range.max}

def getGain(cam):
    range = cam.get_gain_bounds()
    return {"value":cam.get_gain(), "min":range.min, "max":range.max}


def showImg(unusedVar=None):
    global imgShape, lastPictureTime, lastPicture, gStream
    nowtime = time.time()
    buff = gStream.pop_buffer()
    if buff.get_status() != Aravis.BufferStatus.SUCCESS:
        print("missed frame", buff.get_status())
        gStream.push_buffer(buff)
        buff.get_data()
        return lastPicture
    if buff is not None:
        gStream.push_buffer(buff)
        dat = buff.get_data()
        img = np.frombuffer(dat, dtype=np.uint8).reshape(
            imgShape["height"], imgShape["width"], 3)
        lastPictureTime = time.time()
        lastPicture = img
        elapsed = lastPictureTime - nowtime
        # print("showimg elapsed time: ", elapsed)
        return img
    else:
        print("missed frame. no buffer")
    return


def rgb2gray(rgb):
    # r, g = rgb[:, :, 0], rgb[:, :, 1]
    # gray = 0.7 * r + 0.3 * g
    # return gray
    r, g, b = rgb[:, :, 0], rgb[:, :, 1], rgb[:, :, 2]
    gray = 0.8 * r + 0.1 * g + 0.1 * b
    return gray


def findMinMax(mat):
    index = mat.argmax()
    dim = mat.shape
    y = int(index / dim[1])
    x = int(index % dim[1])
    intens = mat[y][x]
    return [x, y, intens]


def brightestPoint(cam, saveLaserDebugImages = False):
    global background
    scale = 3
    try:
        mat1 = showImg(cam)
    except:
        print('skipped frame')
        return {"x": "0", "y": "0", "x1": "0", "y1": "0", "i": "0"}
    mat = rescale(mat1, 1/scale, anti_aliasing=False, channel_axis=-1)
    mat = (mat*255).astype("uint8")
    mat = rgb2gray(mat)

    if background is None:
        background = mat
    fg = np.minimum(np.maximum(mat - background, 0), 255).astype("uint8")
    mm = findMinMax(fg)
    xy = [float(mm[0] + 0.5) / fg.shape[1], float(mm[1] + 0.5) / fg.shape[0]]
    mm[0] = mm[0] * scale
    mm[1] = mm[1] * scale
    background = 0.99 * background + 0.01 * mat
    result = {
        "x": str(mm[0]),
        "y": str(mm[1]),
        "x1": str(xy[0]),
        "y1": str(xy[1]),
        "i": str(mm[2]),
    }
    #
    # Save debug images. Make sure intensity is above 100. Also, only take 5% as to not fill up the hard drive too quickly
    if saveLaserDebugImages and mm[2] > 100 and random.randint(0,100) > 95:
        print('saving')
        saveDebugImages(mat1, fg, result)
    
    return result


def saveDebugImages(original, bg, resultJson):
    mypath = 'tmp/laserDebugImages/'
    filename = f'{mypath}{datetime.datetime.now().timestamp()}'
    try:
        Path(mypath).mkdir(parents=True, exist_ok=True)
        io.imsave(f'{filename}_orig.jpg', original) # original image
        io.imsave(f'{filename}_fg.jpg', bg) # background subtracted image
        augmented = original.copy()
        cx = float(resultJson['x1']) * original.shape[1]
        cy = float(resultJson['y1']) * original.shape[0]
        rr,cc = draw.rectangle_perimeter(start=(cy-20, cx-20), end=(cy+20,cx+20), shape=original.shape)
        augmented[rr,cc] = [255,255,0]
        io.imsave(f'{filename}_aug.jpg', augmented) # augmented with rectange
        with open(f'{filename}_result.json', "w") as text_file: # text file with brightest point
            text_file.write(json.dumps(resultJson, indent=4))
    except Exception as err:
        print('failed to save debug image')
        print(err)




def setDelay(cam, delay):
    global currentDelay, allowTrigger, softwareTriggerFPS
    delay = float(delay)
    changed = False
    if not allowTrigger:
        return changed
    if softwareTriggerFPS > 0:
        return False
    if currentDelay != delay:
        if delay <= 0:
            print("going to auto trigger mode")
            cam.stop_acquisition()
            cam.clear_triggers()
            cam.set_frame_rate(30.0)
            cam.start_acquisition()
            print("going to auto trigger mode")
        else:
            print("going to trigger mode")
            cam.stop_acquisition()
            cam.clear_triggers()
            if cam.get_vendor_name() == 'Allied Vision':
                cam.set_trigger('Line1')
            else:
                cam.set_trigger('Line2')
            print("trigger mode line", cam.get_trigger_source())
            cam.start_acquisition()
        currentDelay = delay
        changed = True
        print("set delay to", delay)
        cleanBuffer(cam)
    return changed


##
# shutter values for the old version were scaled. This on is in microsecods.
#
#
def setShutter(cam, shutter):
    global currentShutter
    shutter = float(10*shutter) # Scaled by ten. This is a holdover from the old version.
    changed = False
    if currentShutter != shutter:
        if shutter <= 0:
            print("\nsetting shutter to auto\n")
            cam.set_exposure_time_auto(Aravis.Auto.CONTINUOUS)
            currentShutter = shutter
            changed = True
            print(getInfo(cam))
        elif shutter > 0:
            # cam.ExposureAuto = "Off"
            bounds = cam.get_exposure_time_bounds()
            clampedShutter = max(min(shutter, bounds.max), bounds.min)
            cam.set_exposure_time_auto(Aravis.Auto.OFF)
            cam.set_exposure_time(clampedShutter)
            currentShutter = clampedShutter
            print("shutter set to:", clampedShutter)
            changed = True
        print("set shutter to", shutter)
    if changed == True:
        cleanBuffer(cam)
    return changed


def setGain(cam, gain):
    gain = float(gain)
    global currentGain
    changed = False
    if currentGain != gain:
        print(currentGain)
        cam.stop_acquisition()
        if gain < 0:
            print("\nsetting gain to auto\n")
            cam.set_gain_auto(Aravis.Auto.CONTINUOUS)
            currentGain = gain
            changed = True
        elif gain >= 0:
            cam.set_gain_auto(Aravis.Auto.OFF)
            bounds = cam.get_gain_bounds()
            gainClamped = max(min(gain, bounds.max), bounds.min)
            cam.set_gain(gainClamped)
            # cam.Gain = gainClamped
            currentGain = float(gainClamped)
            changed = True
            print("set gain to", gain, gainClamped, currentGain)
        cam.start_acquisition()
    if changed == True:
        cleanBuffer(cam)
    return changed


def cleanBuffer(cam, frames=0):
    print('cleaning buffer')
    for i in range(frames):
        showImg(cam)
    # time.sleep(1/30)
    return


def stopCamera(cam):
    cam.stop_acquisition()
    cam.clear_triggers()
    # cam.stop_acquisition()
    # cam.close()  # You should explicitly clean up


cam = 0
