
import camera as cs
import time
from io import BytesIO ## for Python 3
import json
import sys
import os
import signal
import json
from skimage import io

# import cProfile, pstats, io
# from pstats import SortKey

try:
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
    import urlparse
except:
    from http.server import BaseHTTPRequestHandler, HTTPServer
    import urllib.parse as urlparse

myport = 8080
VERSION = 3.1

keepServerRunning = True # set to false to quit
saveDebugLaserImages = False

class GetHandler(BaseHTTPRequestHandler):
    global saveDebugLaserImages
    
    def do_GET(self):
        parsed_path = urlparse.urlparse(self.path)
        realPath = parsed_path.path.lower()
        argums = urlparse.parse_qs( parsed_path.query.lower())
        scriptDir = os.path.dirname(__file__)

        shutter = -999
        gain = -999
        delay = -999
        exposure = -999

        if argums.get('delay'):
            #print( 'delay', argums.get('delay'))
            delay = float( argums.get('delay')[0] )
        if argums.get('shutter'):
            #print( 'shutter', argums.get('shutter'))
            shutter = float( argums.get('shutter')[0])
        if argums.get('gain'):
            #print( 'gain', argums.get('gain'))
            gain = float( argums.get('gain')[0])
        if argums.get('exposure'):
            #print( 'exposure', argums.get('exposure'))
            exposure = float( argums.get('exposure')[0])

        parseArguments( shutter = shutter, gain = gain, delay=delay, exposure=exposure)
        try:
            # print(realPath)
            if 'shot' in realPath or 'image' in realPath:
                self.send_response(200)
                self.send_header('Access-Control-Allow-Origin', '*')
                # self.end_headers()
                if '.png' in realPath:
                    self.send_header('Content-type', 'image/png')
                    self.end_headers()
                    self.wfile.write(getFormattedImage('PNG'))
                else:
                    self.send_header('Content-type', 'image/jpeg')
                    self.end_headers()
                    self.wfile.write(getFormattedImage('JPEG'))
                return
            elif 'getgain' in realPath:
                print("getGain")
                self.send_response(200)
                self.send_header('Access-Control-Allow-Origin', '*')
                self.end_headers()
                jsonVal = json.dumps( cs.getGain(cs.cam) )
                self.wfile.write( jsonVal.encode() )
                return
            elif 'getshutter' in realPath:
                print("getShutter")
                self.send_response(200)
                self.send_header('Access-Control-Allow-Origin', '*')
                self.end_headers()
                jsonVal = json.dumps( cs.getExposureTime(cs.cam) )
                self.wfile.write( jsonVal.encode() )
                return
            elif 'brightestpoint.json' in realPath:
                self.send_response(200)
                self.send_header('Access-Control-Allow-Origin', '*')
                self.end_headers()
                jsn = json.dumps( cs.brightestPoint( cs.cam, saveDebugLaserImages))
                self.wfile.write(jsn.encode())
                return
            # elif 'stream' in realPath or 'mjpg' in realPath: 
            #     self.send_response(200)
            #     self.send_header('Access-Control-Allow-Origin', '*')
            #     self.send_header('Content-type', 'multipart/x-mixed-replace; boundary=frame')
            #     # self.send_header('Cache-Control', 'no-cache, no-store, must-revalidate')
            #     self.end_headers()
            #     count = 0
            #     while True:
            #         self.wfile.write(bytes('--frame\r\n', 'utf8'))
            #         self.wfile.write(bytes('Content-type: image/jpeg\r\n\r\n', 'utf8'))
            #         img = getFormattedImage('JPEG')
            #         self.wfile.write(img)
            #         # print('sent', count)
            #         count = count + 1
            #         time.sleep(1/30) # 30 fps
        except Exception as camError:
            print(camError)
            self.send_response(500)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.end_headers()
            self.wfile.write("Camera failed")
            return
        if 'video.html' in realPath:
            self.send_response(200)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.end_headers()
            f = open( os.path.join(scriptDir,'html/video.html'))
            self.wfile.write(f.read().encode())
            f.close()
        elif 'test.html' in realPath:
            self.send_response(200)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.end_headers()
            f = open( os.path.join(scriptDir,'html/index.html'))
            self.wfile.write(f.read().encode())
            f.close()
        elif 'test.js' in realPath:
            self.send_response(200)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.end_headers()
            f = open( os.path.join(scriptDir,'html/test.js'))
            self.wfile.write(f.read().encode())
            f.close()
        elif 'version.json' in realPath:
            self.send_response(200)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.end_headers()
            jsn = json.dumps( {"version":VERSION })
            self.wfile.write(jsn.encode())
        elif 'tf.html' in realPath:
            self.send_response(200)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.end_headers()
            f = open( u'html/tensorFlowJSWithTreeHouseCamera.html')
            self.wfile.write(f.read().encode())
            f.close()
        elif 'shutdown' in realPath:
            shutdownCameraAndServer(cs.cam, self.server)
            self.send_error(200)
            self.wfile.write("Shutting Down".encode())
        else: #index file
            message_parts = [
                '' + getIndex(),
                '\n\n<pre>CLIENT VALUES:',
                'client_address=%s (%s)' % (self.client_address,
                                            self.address_string()),
                'command=%s' % self.command,
                'path=%s' % self.path,
                'real path=%s' % parsed_path.path,
                'query=%s' % parsed_path.query,
                'request_version=%s' % self.request_version,
                '',
                'SERVER VALUES:',
                'server_version=%s' % self.server_version,
                'sys_version=%s' % self.sys_version,
                'protocol_version=%s' % self.protocol_version,
                '',
                'HEADERS RECEIVED:',
                '</pre>',
                ]
            for name, value in sorted(self.headers.items()):
                message_parts.append('%s=%s' % (name, value.rstrip()))
            message_parts.append('')
            message = '\r\n'.join(message_parts)
            self.send_response(200)
            self.send_header('Access-Control-Allow-Origin', '*')
            self.end_headers()
            self.wfile.write(message.encode())
        return

    def log_message(self, format, *args):
        return


def parseArguments( shutter = -999, gain = -999, delay = -999, exposure=-999):
    shutter = float( shutter)
    gain = float( gain)
    delay = float( delay)
    exposure = float( exposure)
    if( shutter < 0.01):
        shutter = min(0,shutter)
    if( delay < 0.0000001):
        delay = min(0,delay)
    if( gain < 0.01):
        gain = min(0,gain)

    if gain > -999:
        cs.setGain( cs.cam, gain)
    if shutter > -999:
        cs.setShutter( cs.cam, float(shutter) )
    if delay > -999:
        cs.setDelay( cs.cam, delay)

def getFormattedImage(imgFormat):
    outimg = getImage()
    # there is a bug, where the first few images are stale. Ever since upgrading to Aravis 0.8.30 there seem to be stale frames.
    # So throw away 5 images. 
    # TODO: figure out why this is happening and see if it fixes itself in future versions of Aravis > 0.8.30
    for i in range(5):
        outimg = getImage()
    output = BytesIO()
    io.imsave(output, outimg, format=imgFormat)
    message = output.getvalue()
    return message

def getImage():
    cs.lastPictureTime = time.time()
    currimg = cs.showImg(cs.cam)
    #print('image.png taken at', cs.lastPictureTime)
    outimg = currimg.astype('uint8')
    return outimg

def getIndex():
    global myport
    port = str(myport)
    responseString = "<HTML><BODY><h3> You are connected to the python<h1> Web camera server</h1></h3>"
    responseString += "<hr>Usage:<li><b><a href='http://127.0.0.1:" + port + "/shot.jpg'>http://127.0.0.1:" + port + "/shot.jpg</a> </b>"
    responseString += " <li><b><a href='http://127.0.0.1:" + port + "/video.html'>http://127.0.0.1:" + port + "/video.html</a></b> Video Stream Tool"
    # responseString += "<li style='font-size:small'><b><a href='http://127.0.0.1:" + port + "/stream'>stream</a></b> Video Stream (mjpeg)"
    responseString += " <li style='font-size:small'><b><a href='http://127.0.0.1:" + port + "/shot.jpg?shutter=4&gain=0'>http://127.0.0.1:" + port + "/shot.jpg?shutter=4&gain=0</a></b>. Shutter and Gain can be set absolutely"
    responseString += " <li style='font-size:small'><b><a href='http://127.0.0.1:" + port + "/shot.jpg?delay=0.002'>http://127.0.0.1:" + port + "/shot.jpg?delay=0.002</a></b>. The delay in seconds afer GPIO trigger signal"
    responseString += " <li style='font-size:small'><b><a href='http://127.0.0.1:" + port + "/shot.jpg?shutter=-1&gain=-1&delay=-1'>http://127.0.0.1:" + port + "/shot.jpg?shutter=-1&gain=-1&delay=-1</a></b>. Negative numbers turn on automatic mode"
    responseString += " <li style='font-size:small'><b><a href='http://127.0.0.1:" + port + "/?shutter=4&gain=0&delay=0.002'>http://127.0.0.1:" + port + "/?shutter=4&gain=0&delay=0.002</a></b>. You don't need to take a picture to set the parameters "
    responseString += " <li style='font-size:small'><b><a href='http://127.0.0.1:" + port + "/shot.png'>http://127.0.0.1:" + port + "/shot.png</a></b>. Different formats based on extension. Supported extensions are jpg, png, and bmp."
    responseString += " <li style='font-size:small'><b><a href='http://127.0.0.1:" + port + "/brightestpoint.json'>http://127.0.0.1:" + port + "/brightestpoint.json</a></b>. Find the brightest Point"
    responseString += " <li style='font-size:small'><b><a href='http://127.0.0.1:" + port + "/brightestpoint.json?shutter=40&gain=0'>http://127.0.0.1:" + port + "/brightestpoint.json?shutter=40&gain=0</a></b>. Find the brightest Point"
    responseString += " <li style='font-size:small'><b><a href='http://127.0.0.1:" + port + "/test.html'>http://127.0.0.1:" + port + "/stressTest.html</a></b>... more tests ..."
    responseString += " <li style='font-size:small'><b><a href='http://127.0.0.1:" + port + "/tf.html'>http://127.0.0.1:" + port + "/tf.html</a></b>tensor flow test"
    responseString += "<br><hr><h2>Camera Information</h2><pre>" + cs.getInfo(cs.cam) +"</pre>"
    responseString += "</BODY></HTML>"
    return responseString

def defaultStatusCallback(status):
    print(status)

def shutdownCameraAndServer(cam, server):
    global keepServerRunning
    cs.stopCamera(cam)
    keepServerRunning=False
    print('camera stopped')




def main(noTrigger=False):
    global keepServerRunning
    global saveDebugLaserImages

    print ('Starting server, use <Ctrl-C> to stop')
    server = HTTPServer(('0.0.0.0', myport), GetHandler)

    def sigterm_handler(_signo, _stack_frame):
        print('exiting')
        sys.exit(0)
    signal.signal(signal.SIGTERM, sigterm_handler)

    print ('port', myport)
    if noTrigger or "--notrigger" in sys.argv or "--no-trigger" in sys.argv:
        print("NO TRIGGER")
        cs.allowTrigger = False
    if "--softwaretrigger" in sys.argv or "--software-trigger" in sys.argv:
        print("SOFTWARE TRIGGER")
        cs.softwareTriggerFPS = 30
    
    if "--laser-debug" in sys.argv or "--laserdebug" in sys.argv:
        print("Laser Debug mode on")
        saveDebugLaserImages = True

    print('starting camera')
    cs.initCam()
    try:
        while keepServerRunning:
            server.handle_request()
    except:
        print('something happened', sys.exc_info()[0])
    print('Exiting camera server')
    shutdownCameraAndServer(cs.cam, server)


if __name__ == '__main__':
    print('hello')
    main()
    # with cProfile.Profile() as pr:
        # main()
    # s = io.StringIO()
    # sortby = SortKey.CUMULATIVE
    # ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    # ps.print_stats(120)
    # print(s.getvalue())

 
